import click
import os
import os.path
import sys
import shutil
import yaml
import random
import string
import itertools
import time
import platform
from .proatacHelp import *

# ----------------------------------
# Core object for handling the .yaml
# ----------------------------------

class proatacProject():
	def __init__(self, yaml, script_dir):
		
		# Basic attributes
		self.yaml = yaml
		self.project_name = self.yaml['project_name']
		self.project_dir = self.yaml['project_dir']
		self.analysis_person = self.yaml['analyst']
		
		if((' ' in self.yaml['project_name']) or (' ' in self.yaml['project_dir'])):
			sys.exit("ERROR: remove white space from project_name and project_dir variables in .yaml")
		
		# ------------------------------
		# Process reference genome stuff
		# ------------------------------
		
		# Computing configuration
		if ("read_quality" in self.yaml['parameters']) and (str(self.yaml['parameters']['read_quality']) != "None"):
			if(int(self.yaml['parameters']['read_quality']) < 0):
				sys.exit("ERROR: read_quality as specified in the .yaml seems fishy; make sure it's a positive integer")
			self.read_quality = str(int(self.yaml['parameters']['read_quality']))
		else:
			self.read_quality = str(30)

		# Computing configuration
		if ("max_cores" in self.yaml['parameters']) and (str(self.yaml['parameters']['max_cores']) != "None"):
			self.max_cores = str(self.yaml['parameters']['max_cores'])
		else:
			self.max_cores = str(2)
		
		# Computing configuration
		if ("extract_mito" in self.yaml['parameters']) and (str(self.yaml['parameters']['extract_mito']) != "None"):
			self.extract_mito = str(self.yaml['parameters']['extract_mito'])
		else:
			self.extract_mito = "false"

		# Computing configuration
		if ("chr_name_length" in self.yaml['parameters']) and (str(self.yaml['parameters']['chr_name_length']) != "None"):
			self.chr_name_length = str(self.yaml['parameters']['chr_name_length'])
		else:
			self.chr_name_length = str(10)
		
		# Figure out operating system
		self.os = "linux"
		if(platform.platform()[0:5]=="Darwi"):
			self.os = "mac"
		
		if(self.os == "mac"):
			self.peat_path = script_dir + "/bin/mac/PEAT_cl123_mac"
			self.pigz_path = script_dir + "/bin/mac/pigz_mac"
		else:
			self.peat_path = script_dir + "/bin/linux/PEAT_cl123_linux"
			self.pigz_path = script_dir + "/bin/linux/pigz_linux"
		
		outfolder = os.path.abspath(yaml['project_dir']) 
		logfolder = outfolder + "/logs"
		logf = open(logfolder + "/base.proatac.log", 'a')
		
		# ------------------------------
		# Process peak information
		# ------------------------------
		
		self.reference_genome = self.yaml['reference_genome']
		supported_genomes = ['hg19', 'hg38', 'mm9', 'mm10', 'hg19_mm10_c']
		if any(self.reference_genome in s for s in supported_genomes):
			click.echo(gettime() + "Found designated reference genome: %s" % self.reference_genome, logf)
			self.tssFile = script_dir + "/anno/TSS/" + self.reference_genome + ".refGene.TSS.bed"
			self.blacklistFile = script_dir + "/anno/blacklist/" + self.reference_genome + ".full.blacklist.bed"
			self.bedtoolsGenomeFile = script_dir + "/anno/bedtools/chrom_" + self.reference_genome + ".sizes"
			
			# Set up effective genome size for macs2
			if self.reference_genome == 'hg19':
				self.BSgenome = 'BSgenome.Hsapiens.UCSC.hg19'
				self.macs2_genome_size = 'hs'
			elif self.reference_genome == 'hg38':
				self.BSgenome = 'BSgenome.Hsapiens.UCSC.hg38'
				self.macs2_genome_size = 'hs'
			elif self.reference_genome == 'mm9':
				self.BSgenome = 'BSgenome.Mmusculus.UCSC.mm9'
				self.macs2_genome_size = 'mm'
			elif self.reference_genome == 'mm10':
				self.BSgenome = 'BSgenome.Mmusculus.UCSC.mm10'
				self.macs2_genome_size = 'mm'
			else:
				self.BSgenome = ''
				self.macs2_genome_size = '4.57e9'
		else: 
			click.echo(gettime() + "Could not identify this reference genome: %s" % self.reference_genome, logf)
				
		if ("peak_settings" in self.yaml):
			if "tss_file" in self.yaml['peak_settings']:
				b = self.yaml['peak_settings']['tss_file']
				if(b != ''):
					self.tssFile = os.path.realpath(b)
			if "blacklist_file" in self.yaml['peak_settings']:
				b = self.yaml['peak_settings']['blacklist_file']
				if(b != ''):
					self.blacklistFile = os.path.realpath(b)
			if "bedtools_genome" in self.yaml['peak_settings']:
				b = self.yaml['peak_settings']['bedtools_genome']
				if(b != ''):
					self.bedtoolsGenomeFile = os.path.realpath(b)
			if "bs_genome" in self.yaml['peak_settings']:
				b = self.yaml['peak_settings']['bs_genome']
				if(b != ''):
					self.bsGenome = b
			if "individual_peaks" in self.yaml['peak_settings']:
				b = self.yaml['peak_settings']['individual_peaks']
				if(b != ''):
					self.individual_peaks = b
			if "n_peaks" in self.yaml['peak_settings']:
				b = self.yaml['peak_settings']['n_peaks']
				if(b != ''):
					self.n_peaks = b
			if "peak_width" in self.yaml['peak_settings']:
				b = self.yaml['peak_settings']['peak_width']
				if(b != ''):
					self.peak_width = b

		# ------------------------
		# Process dependency paths
		# ------------------------

		# bedtools
		if(self.yaml['paths']['bedtools_path'] != ''):
			self.java_path = self.yaml['paths']['bedtools_path']
		else:
			self.bedtools_path = shutil.which("bedtools")
		if(str(self.bedtools_path) == "None"):
			sys.exit("ERROR: cannot find bedtools in environment; set the 'bedtools_path' in the .yaml file or add to PATH")

		# bowtie2
		if(self.yaml['paths']['bowtie2_path'] != ''):
			self.bowtie2_path = self.yaml['paths']['bowtie2_path']
		else:
			self.bowtie2_path = shutil.which("bowtie2")
		if(str(self.bowtie2_path) == "None"):
			sys.exit("ERROR: cannot find bowtie2 in environment; set the 'bowtie2_path' in the .yaml file or add to PATH")
		
		# bowtie2 index
		bwt2idxfiles = os.popen("ls " + self.yaml['paths']['bowtie2_index']+ "*.bt2").read().strip().split("\n")
		if(len(bwt2idxfiles) < 6):
			sys.exit("ERROR: cannot find bowtie2 index; make sure to add the prefix along with the folder path")
		else:
			self.bowtie2_index = self.yaml['paths']['bowtie2_index']
		
		# macs2	
		if(self.yaml['paths']['macs2_path'] != ''):
			self.macs2_path = self.yaml['paths']['macs2_path']
		else:
			self.macs2_path = shutil.which("macs2")
		if(str(self.macs2_path) == "None"):
			sys.exit("ERROR: cannot find macs2 in environment; set the 'macs2_path' in the .yaml file or add to PATH")
		
		# samtools
		if(self.yaml['paths']['samtools_path'] != ''):
			self.samtools_path = self.yaml['paths']['samtools_path']
		else:
			self.samtools_path = shutil.which("samtools")
		if(str(self.samtools_path) == "None"):
			sys.exit("ERROR: cannot find samtools in environment; set the 'samtools_path' in the .yaml file or add to PATH")
				
		# R
		if(self.yaml['paths']['R_path'] != ''):
			self.R_path = self.yaml['paths']['R_path']
		else:
			self.R_path = shutil.which("R")
		if(str(self.R_path) == "None"):
			sys.exit("ERROR: cannot find R in environment; set the 'R_path' in the .yaml file or add to PATH")

		# Java
		if(self.yaml['paths']['java_path'] != ''):
			self.java_path = self.yaml['paths']['java_path']
		else:
			self.java_path = shutil.which("java")
		if(str(self.java_path) == "None"):
			sys.exit("ERROR: cannot find Java in environment; set the 'java_path' in the .yaml file or add to PATH")
										
		# Check for R package dependencies
		check_R_packages(['ggplot2', 'tidyverse'], self.R_path)
		
		# The final step should be fast, so remove the file that coordinates all samples if it exists
		listAllSamples = outfolder + '/.internal/parseltongue/allsamples.csv'
		if os.path.exists(listAllSamples):
			os.remove(listAllSamples)
		
		# Process sequencing directories		
		for run in self.yaml['sequencing_directories']:
			process_seq_dir(run, logf, listAllSamples)
		
		# Check to see if any sample names are duplicated
		with open(listAllSamples) as f:
			seen = set()
			for line in f:
				line_lower = line.split(",")[0]
				if line_lower in seen:
					sys.exit("ERROR: found multiple sample IDs specified the same way; this will cause problems down the road; quitting now")
				else:
					seen.add(line_lower)
		
		logf.close()



import yaml
import itertools
import time
import re
import os
import sys
import csv

def string_hamming_distance(str1, str2):
    """
    Fast hamming distance over 2 strings known to be of same length.
    In information theory, the Hamming distance between two strings of equal 
    length is the number of positions at which the corresponding symbols 
    are different.
    eg "karolin" and "kathrin" is 3.
    """
    return sum(itertools.imap(operator.ne, str1, str2))


def rev_comp(seq):
    """
    Fast Reverse Compliment
    """  
    tbl = {'A':'T', 'T':'A', 'C':'G', 'G':'C', 'N':'N'}
    return ''.join(tbl[s] for s in seq[::-1])

def parse_manifest(manifest):
	"""
	Basic function to parse yaml/yml
	"""
	samples = []
	if manifest.endswith(('.yaml', '.yml')):
		with open(manifest, 'r') as f: 
			m = yaml.load(f)
		return m
	else:
		click.echo(gettime() + "Please specify a valid .yaml file for analysis")

def gettime(): 
	"""
	Matches `date` in Linux
	"""
	return(time.strftime("%a ") + time.strftime("%b ") + time.strftime("%d ") + time.strftime("%X ") + 
		time.strftime("%Z ") + time.strftime("%Y")+ ": ")
		

def findIdx(list1, list2):
	"""
	Return the indices of list1 in list2
	"""
	return [i for i, x in enumerate(list1) if x in list2]

def check_R_packages(required_packages, R_path):
	"""
	Determines whether or not R packages are properly installed
	"""
	installed_packages = os.popen(R_path + ''' -e "installed.packages()" | awk '{print $1}' | sort | uniq''').read().strip().split("\n")
	if(not set(required_packages) < set(installed_packages)):
		sys.exit("ERROR: cannot find the following R package: " + str(set(required_packages) - set(installed_packages)) + "\n" + 
			"Install it in your R console and then try rerunning proatac (but there may be other missing dependencies).")
		

def process_seq_dir(d, logf, listAllSamples):
	"""
	Function that takes a dictionary parsed from the main .yaml
	and returns something more coherent to be processed downstream
	Could parameterize this function further given more versions of scATAC
	that would be worth processing.
	"""
	
	name = d['name']
	version = d['version']
	directory =  d['dir']
	fastq_path = d['fastq_path']
	
	# Handle whether the reads are R1/R2 (default) or something else
	# Some attempt on the author's part to be savvy here. 
	if 'read' in d:
		read = d['read']
	else:
		read = ["R1", "R2"]
	if(len(read) != 2):
		sys.exit("ERROR: invalid input-- " + read + " -- for 'read' in .yaml; need length 2 python vector")
	
	# Go get the fastq files
	noSID = re.sub("{sample_id}", "", fastq_path)
	read1 = re.sub("{read}", read[0], noSID)
	read2 = re.sub("{read}", read[1], noSID)
	
	reads1 = os.popen("ls " + directory + "/*" + read1).read().strip().split("\n")
	reads2 = os.popen("ls " + directory + "/*" + read2).read().strip().split("\n")
	
	# Parse out the sample names
	temp1 = [re.sub(read1, "", x) for x in reads1]
	samples = [re.sub(directory + "/", "", x) for x in temp1]
	
	# Remove / filter samples as requested if applicable; also remove those from the file listing
	if 'keep_samples' in d:
		keep_samples = d['keep_samples']
	else:
		keep_samples = ""
		
	if(keep_samples != ""):
		keeplist = findIdx(samples, keep_samples) # indicies of samples to be kept
		samples = [samples[i] for i in keeplist]
		reads1 = [reads1[i] for i in keeplist]
		reads2 = [reads2[i] for i in keeplist]
	
	if 'remove_samples' in d:
		remove_samples = d['remove_samples']
	else:
		remove_samples = ""
		
	if(remove_samples != ""):
		rmlist = findIdx(samples, remove_samples) # indices of samples that should be removed
		allidx = range(len(samples)) # 1:n
		keeplist2 = [x for x in allidx if x not in rmlist] # indices in 1:n not in list of indices to be removed
		samples = [samples[i] for i in keeplist2]
		reads1 = [reads1[i] for i in keeplist2]
		reads2 = [reads2[i] for i in keeplist2]
	
	with open(listAllSamples, 'a') as f:
		writer = csv.writer(f)
		rows = zip(samples,reads1,reads2)
		for row in rows:
			writer.writerow(row)
	
	


import click
import os
import os.path
import sys
import shutil
import yaml
import random
import string
import itertools
import time
from pkg_resources import get_distribution
from subprocess import call, check_call
from .proatacHelp import *
from .proatacProjectClass import *



# ------------------------------
# Command line arguments
# ------------------------------	

@click.command()
@click.version_option()
@click.option('--check', is_flag=True, help='[MODE] Check to see if all dependencies are properly configured.')
@click.option('--stingy', is_flag=True, help='Space-efficient analyses; remove non-vital intermediate files.')
@click.argument('manifest')


def main(manifest, check, stingy):
	"""Preprocessing ATAC and scATAC Data."""
	__version__ = get_distribution('proatac').version
	script_dir = os.path.dirname(os.path.realpath(__file__))

	click.echo(gettime() + "Starting proatac pipeline v%s" % __version__)
	ymml = parse_manifest(manifest)

	
	# -------------------------------
	# Utility functions and variables
	# -------------------------------
	
	outfolder = os.path.abspath(ymml['project_dir']) 
	logfolder = outfolder + "/logs"
	internfolder = outfolder + "/.internal"
	parselfolder = internfolder + "/parseltongue"
	
	# Check if directories exist; make if not
	if not os.path.exists(outfolder):
		os.makedirs(outfolder)
	if not os.path.exists(logfolder):
		os.makedirs(logfolder)	
	if not os.path.exists(internfolder):
		os.makedirs(internfolder)
		with open(internfolder + "/README" , 'w') as outfile:
			outfile.write("This folder creates important (small) intermediate; don't modify it.\n\n")	
	if not os.path.exists(parselfolder):
		os.makedirs(parselfolder)
		with open(parselfolder + "/README" , 'w') as outfile:
			outfile.write("This folder creates intermediate output to be interpreted by Snakemake; don't modify it.\n\n")	
	cwd = os.getcwd()
	logf = open(logfolder + "/base.proatac.log", 'a')
		
	# Main Project Variable; output to .yaml so it's clear what's going on
	p = proatacProject(ymml, script_dir)
	projectdict = p.__dict__
	del projectdict["yaml"]
	
	# Copy original yaml
	cpyaml = 'cp ' + manifest + ' ' + logfolder +  "/supplied.yaml"
	os.system(cpyaml)
	
	y0 = logfolder + "/interpreted.params.yaml"
	with open(y0, 'w') as yaml_file:
		yaml.dump(projectdict, yaml_file, default_flow_style=False)
	
	# -------------------------------
	# Atypical analysis modes
	# -------------------------------	
	
	if check:
		sys.exit("Success! We're reasonably confident that all dependencies and files are good to go!")

	# -----------------------------------
	# ACTUAL PREPROCESSING / SNAKE MAKING
	# -----------------------------------
	
	click.echo(gettime() + "Project .yaml successfully loaded. ", logf)
	
	# -------------
	# Adapter Trim
	# -------------
	trimfolder = outfolder + "/01_trimmed"
	if not os.path.exists(trimfolder + "_reads"):
		os.makedirs(trimfolder+ "_reads")
				
	click.echo(gettime() + "Trimming samples", logf)
	
	snakedict1 = {'allsamples' : parselfolder + "/allsamples.csv", 'outdir' : outfolder,
		'scriptdir' : script_dir, 'PEAT' : p.peat_path, 'pigz' : p.pigz_path}
		
	y1 = parselfolder + "/snake.trim.yaml"
	with open(y1, 'w') as yaml_file:
		yaml.dump(snakedict1, yaml_file, default_flow_style=False)
		
	snakecall1 = 'snakemake --snakefile ' + script_dir + '/bin/snake/Snakefile.Trim --cores ' + p.max_cores + ' --config cfp="' + y1 + '"'
	os.system(snakecall1)
	click.echo(gettime() + "Sample trimming done.", logf)
	
	# -------------
	# Alignment
	# -------------
	if not os.path.exists(logfolder + "/bowtie2logs"):
		os.makedirs(logfolder+ "/bowtie2logs")
	if not os.path.exists(outfolder + "/02_aligned_reads"):
		os.makedirs(outfolder+ "/02_aligned_reads")
		
	click.echo(gettime() + "Aligning samples", logf)
	
	snakedict2 = {'bowtie2' : p.bowtie2_path, 'bowtie2index' : p.bowtie2_index,
		'outdir' : outfolder, 'samtools' : p.samtools_path}
	
	y2 = parselfolder + "/snake.align.yaml"
	with open(y2, 'w') as yaml_file:
		yaml.dump(snakedict2, yaml_file, default_flow_style=False)
	
	snakecall2 = 'snakemake --snakefile ' + script_dir + '/bin/snake/Snakefile.Align --cores ' + p.max_cores + ' --config cfp="' + y2 + '"'
	os.system(snakecall2)
	
	# -------------
	# bam process
	# -------------
	if not os.path.exists(outfolder + "/03_processed_reads"):
		os.makedirs(outfolder+ "/03_processed_reads")
	if not os.path.exists(outfolder + "/03_processed_reads/individual"):
		os.makedirs(outfolder+ "/03_processed_reads/individual")
	if not os.path.exists(logfolder + "/rmduplogs"):
		os.makedirs(logfolder+ "/rmduplogs")
		
	click.echo(gettime() + "Cleaning up .bam files", logf)
	
	# Determine chromosomes to keep / filter
	chrs = os.popen(p.samtools_path + " idxstats " +  outfolder + "/02_aligned_reads/* | cut -f1").read().strip().split("\n")
	rmchrlist = ["*", "chrY", "MT", "chrM"]
	keepchrs = [x for x in chrs if x not in rmchrlist and len(x) < int(p.chr_name_length)]
	
	snakedict3 = {'keepchrs' : keepchrs, 'read_quality' : p.read_quality, 'java' : p.java_path,
		'outdir' : outfolder, 'samtools' : p.samtools_path, 'project_name' : p.project_name, 'scriptdir' : script_dir}
		
	y3 = parselfolder + "/snake.bamprocess.yaml"
	with open(y3, 'w') as yaml_file:
		yaml.dump(snakedict3, yaml_file, default_flow_style=False)
	
	snakecall3 = 'snakemake --snakefile ' + script_dir + '/bin/snake/Snakefile.BamProcess --cores ' + p.max_cores + ' --config cfp="' + y3 + '"'
	os.system(snakecall3)
	
	# ---------------------
	# Mitochondria
	# ---------------------
	if(p.extract_mito):
		click.echo(gettime() + "Extracting mitochondrial reads", logf)
		
		if not os.path.exists(outfolder + "/03_processed_reads/mito"):
			os.makedirs(outfolder+ "/03_processed_reads/mito")
			
		# Determine mitochondrial chromosomes
		mitochrs = []
		for name in chrs:
			if 'MT' in name or 'chrM' in name:
				mitochrs.append(name)
				
		snakedictM = {'mitochrs' : mitochrs, 'outdir' : outfolder, 'samtools' : p.samtools_path, 
			'project_name' : p.project_name}
		yM = parselfolder + "/snake.mito.yaml"

		with open(yM, 'w') as yaml_file:
			yaml.dump(snakedictM, yaml_file, default_flow_style=False)
		
		snakecallM = 'snakemake --snakefile ' + script_dir + '/bin/snake/Snakefile.MitoProcess --cores ' + p.max_cores + ' --config cfp="' + yM + '"'
		os.system(snakecallM)
		
		
	# ---------------------
	# Peaks
	# ---------------------
	if not os.path.exists(outfolder + "/04_qc"):
		os.makedirs(outfolder+ "/04_qc")
		
	snakedict4 = {
		'bedtools' : p.bedtools_path, 'blacklistFile' : p.blacklistFile, 'macs2' : p.macs2_path,
		'macs2_genome_size' : p.macs2_genome_size, 'n_peaks' : p.n_peaks, 'outdir' : outfolder,
		'peak_width': p.peak_width, 'project_name' : p.project_name, 'R' : p.R_path, 'script_dir' : script_dir
	}	
		
	y4 = parselfolder + "/snake.callpeaksone.yaml"
	with open(y4, 'w') as yaml_file:
		yaml.dump(snakedict4, yaml_file, default_flow_style=False)
	
	snakecall4 = 'snakemake --snakefile ' + script_dir + '/bin/snake/Snakefile.CallPeaksOne --cores ' + p.max_cores + ' --config cfp="' + y4 + '"'
	os.system(snakecall4)
	
	# ---------------------
	# Individual QC
	# ---------------------	
	if not os.path.exists(outfolder + "/04_qc/individualQC"):
		os.makedirs(outfolder+ "/04_qc/individualQC")
		
	snakedict5 = {
		'bedtools' : p.bedtools_path, 'outdir' : outfolder,
		'project_name' : p.project_name, 'R' : p.R_path, 'samtools' : p.samtools_path, 
		'script_dir' : script_dir
	}	
		
	y5 = parselfolder + "/snake.qcstats.yaml"
	with open(y5, 'w') as yaml_file:
		yaml.dump(snakedict5, yaml_file, default_flow_style=False)
	
	snakecall5 = 'snakemake --snakefile ' + script_dir + '/bin/snake/Snakefile.QCstats --cores ' + p.max_cores + ' --config cfp="' + y5 + '"'
	os.system(snakecall5)
		
	# Suspend logging
	logf.close()
	
	
